#!/usr/bin/env python

# This code is part of the 'enjuewemela' game
# License: GPLv3
# Main author: Facundo Batista
# Code, bug tracker, etc:
#   https://launchpad.net/enjuewemela/
#
"""Main program for the game."""

from __future__ import division

import random
import time

import cocos
from cocos.director import director
from cocos import actions, scene
from cocos.text import Label

import pyglet
from pyglet.window import key as keyboard

import logic_board
import stats
from gamemenu import MainMenu
from gems import gems
from config import (
    ARISTA, BASE_COLOR, BORDER, PIECE_TRIP_TIME, FALL_TIME,
)
from sound import sounds


class BackgroundLayer(cocos.layer.Layer):
    """Background layer for all the game."""

    def __init__(self):
        super(BackgroundLayer, self).__init__()
        self.img = pyglet.resource.image('background_image.png')

    def draw(self):
        """Called when the layer is drawn."""
        self.img.blit(0, 0)


class Board(cocos.draw.Canvas):
    """Static layer for the board."""

    def __init__(self):
        cocos.draw.Canvas.__init__(self)

        win_w, win_h = director.get_window_size()
        self.ini_x = (win_w - ARISTA * 8) // 2
        self.ini_y = (win_h - ARISTA * 8) // 2
        self.center = (win_w // 2, win_h // 2)

    def render(self):
        """Render the board."""
        ini_x, ini_y = self.ini_x, self.ini_y
        self.set_endcap(cocos.draw.BUTT_CAP)

        # big white base
        self.set_stroke_width(ARISTA*8)
        self.set_color((255, 255, 255, 140))
        self.move_to((ini_x + ARISTA*4, ini_y))
        self.line_to((ini_x + ARISTA*4, ini_y + ARISTA*8))

        # two colors for the slots
        color1 = BASE_COLOR + (200,)
        color2 = BASE_COLOR + (100,)

        # slots
        self.set_stroke_width(ARISTA)
        base_x = ini_x + ARISTA // 2   # vertical paint strokes
        for x in range(8):
            for y in range(8):
                if (-1) ** (x+y) == -1:
                    self.set_color(color1)
                else:
                    self.set_color(color2)

                pos_move_to = (base_x + ARISTA*x, ini_y + ARISTA*y)
                self.move_to(pos_move_to)

                pos_line_to = (base_x + ARISTA*x, ini_y + ARISTA*(y+1))
                self.line_to(pos_line_to)

        # border!
        self.set_stroke_width(BORDER)
        board = ARISTA * 8
        self.set_color(BASE_COLOR + (255,))

        # border, bottom line
        self.move_to((ini_x - BORDER,         ini_y - BORDER // 2))
        self.line_to((ini_x + BORDER + board, ini_y - BORDER // 2))

        # border, left line
        self.move_to((ini_x - BORDER // 2, ini_y - BORDER))
        self.line_to((ini_x - BORDER // 2, ini_y + BORDER + board))

        # border, right line
        self.move_to((ini_x + board + BORDER // 2, ini_y - BORDER))
        self.line_to((ini_x + board + BORDER // 2, ini_y + BORDER + board))

        # border, top line
        self.move_to((ini_x - BORDER,         ini_y + board + BORDER // 2))
        self.line_to((ini_x + BORDER + board, ini_y + board + BORDER // 2))

    def get_slot_center(self, pos_x, pos_y):
        """Return the coordinates of the center of the given slot."""
        x = self.ini_x + ARISTA * pos_x + ARISTA // 2
        y = self.ini_y + ARISTA * pos_y + ARISTA // 2
        return x, y

    def get_slot_from_cords(self, coord_x, coord_y):
        """Return the slot for given coordinates."""
        bx = coord_x - self.ini_x
        pos_x = bx // ARISTA
        if not (0 <= pos_x < 8):
            return None

        by = coord_y - self.ini_y
        pos_y = by // ARISTA
        if not (0 <= pos_y < 8):
            return None

        return int(pos_x), int(pos_y)

    def get_cords_from_slot(self, pos_x, pos_y):
        """Return the coordinates for given slot."""
        x = self.ini_x + pos_x * ARISTA
        y = self.ini_y + pos_y * ARISTA
        return x, y


class PieceHolder(object):
    """Class to handle pieces."""

    def __init__(self):
        self.slot2piece = {}
        self.piece2slot = {}

    def __str__(self):
        """Shows itself as a board."""
        return str(self.get_board())

    def set_piece(self, piece, slot_x, slot_y):
        """Set a piece in a slot."""
        self.piece2slot[piece] = slot_x, slot_y
        self.slot2piece[slot_x, slot_y] = piece

    def remove_piece(self, slot_x, slot_y):
        """Remove a piece from a slot."""
        piece = self.slot2piece.pop((slot_x, slot_y))
        del self.piece2slot[piece]
        return piece

    def get_piece(self, slot_x, slot_y):
        """Return the piece in the slot, None if any."""
        try:
            return self.slot2piece[slot_x, slot_y]
        except KeyError:
            return None

    def get_board(self):
        """Return the complete board."""
        board = []
        for y in range(8):
            row = []
            for x in range(8):
                try:
                    piece = self.slot2piece[x, 7-y]
                except KeyError:
                    row.append(None)
                else:
                    row.append(piece.pos_id)
            board.append(row)
        return logic_board.FakeBoard(board)

    def check_matches(self):
        """Tell in which positions there're matches."""
        return logic_board.detect(self.get_board())

    def future_guess(self):
        """Guess which piece the user could move to have a match."""
        return logic_board.future_guess(self.get_board())

    def piece_fall(self, piece, steps):
        """Make the piece fall."""
        # piece to slot
        slot_x, old_slot_y = self.piece2slot[piece]
        new_slot_y = old_slot_y - steps
        self.piece2slot[piece] = slot_x, new_slot_y

        # slot to piece
        del self.slot2piece[slot_x, old_slot_y]
        self.slot2piece[slot_x, new_slot_y] = piece

    def try_swap(self, pc1, pc2):
        """Try to swap the pieces."""
        # get slots
        sl1 = self.piece2slot[pc1]
        sl2 = self.piece2slot[pc2]

        # swap and test
        self.piece2slot[pc1] = sl2
        self.piece2slot[pc2] = sl1
        self.slot2piece[sl1] = pc2
        self.slot2piece[sl2] = pc1
        m = self.check_matches()

        if not m:
            # oops, invalid move, go back!
            self.piece2slot[pc1] = sl1
            self.piece2slot[pc2] = sl2
            self.slot2piece[sl1] = pc1
            self.slot2piece[sl2] = pc2
        return m


class PieceManager(cocos.layer.Layer):
    """Manager class for the gems."""

    def __init__(self, board, st):
        super(PieceManager, self).__init__()
        self.board = board
        self.game_stats = st
        self.ph = PieceHolder()
        self.falling = False
        self.swapping = False
        self.ended = False
        self.paused = False

        init_scenario = logic_board.generate_nonmatch_board()
        for x in range(8):
            for y in range(8):
                which = init_scenario[x][y]
                gem = gems[which]
                self.game_stats.add_piece(gem)

                real_cords =  self.board.get_slot_center(x, y)
                gem.position = real_cords
                gem.orig_pos = real_cords
                self.ph.set_piece(gem, x, y)
                self.add(gem)
        self.game_stats.redraw_stats()

        # pause label
        self.pause_label = Label(_("(game paused)"), board.center,
                                 font_name='Sansation',
                                 font_size=32, color=(30, 30, 30, 255),
                                 anchor_x='center', anchor_y='center')

    def set_pause(self, now_paused):
        """Sets or unsets the pause."""
        self.paused = now_paused
        pieces = self.ph.piece2slot.keys()
        if now_paused:
            new_pieces_opacity = 0
            self.add(self.pause_label)
        else:
            new_pieces_opacity = 255
            self.remove(self.pause_label)
        for p in pieces:
            p.opacity = new_pieces_opacity

    def hand_piece_select(self, piece):
        """Select a gem."""
        piece.scale *= 1.2

    def hand_piece_move(self, piece, coords):
        """Move the gem."""
        piece.position = coords

    def hand_piece_reset(self, piece):
        """Reset the gem movement."""
        piece.position = piece.orig_pos
        piece.scale = gems.resize_factor

    def hand_piece_switch(self, from_piece, to_piece):
        """Switch two gems."""
        self.swapping = True

        # move the selected one
        from_piece.do(actions.MoveTo(to_piece.orig_pos, PIECE_TRIP_TIME))

        # shrink the new one, and move it
        to_piece.scale /= 1.2
        to_piece.do(actions.MoveTo(from_piece.orig_pos, PIECE_TRIP_TIME))

        # schedule finish
        self.call_later(PIECE_TRIP_TIME, self._end_move, from_piece, to_piece)

    def _end_move(self, pc1, pc2):
        """Finish the move."""
        # we'll use the user move after finishing with board filling
        self.swapping = False
        if self.falling:
            print "waiting!"
            self.call_later(.2, self._end_move, pc1, pc2)
            return
        self.falling = True

        # swap and check match
        m = self.ph.try_swap(pc1, pc2)

        if m:
            # yes! confirm the move
            # restart scales and swap their positions...
            pc1.scale = gems.resize_factor
            pc2.scale = gems.resize_factor
            pc1.orig_pos, pc2.orig_pos = pc2.orig_pos, pc1.orig_pos

            # score! score! score!
            self._matches(m, from_user=True)
        else:
            # move back the pieces
            backmove_time = PIECE_TRIP_TIME * .7
            pc1.do(actions.MoveTo(pc1.orig_pos, backmove_time))
            pc2.do(actions.MoveTo(pc2.orig_pos, backmove_time))

            def end_backmove():
                """Finish the back move."""
                self.hand_piece_reset(pc1)
                self.hand_piece_reset(pc2)
                self.falling = False
            self.call_later(backmove_time, end_backmove)

    def _matches(self, matches, from_user):
        """There's a match, remove the pieces and fill."""
        # pumba!
        sounds.play("pumba")

        # blow up the pieces
        to_blow = set()
        for pack in matches:
            for pair in pack:
                to_blow.add(pair)

        # fill the board again
        for x in range(8):
            empties = 0
            for y in range(8):
                p = self.ph.get_piece(x, y)
                if (x, y) in to_blow:
                    empties += 1
                    self.remove(p)
                    self.ph.remove_piece(x, y)
                    self.game_stats.destroy_piece(p)
                else:
                    if empties:
                        self._let_it_fall(p, empties)

            # fill the column
            for i in range(empties):
                upper_slot_y = 8 + i
                coord_x, coord_y =  self.board.get_slot_center(x, upper_slot_y)
                newpiece = gems[random.randint(0, 6)]
                newpiece.orig_pos = coord_x, coord_y
                newpiece.position = coord_x, coord_y
                self.ph.set_piece(newpiece, x, upper_slot_y)
                self.game_stats.add_piece(newpiece)
                self.add(newpiece)
                self._let_it_fall(newpiece, empties)

        # fill the match position and keep score
        pos_match = []
        for m in matches:
            slot = logic_board.get_center(m)
            coords = self.board.get_slot_center(*slot)
            pos_match.append((coords, m))
        self.game_stats.matches(pos_match, from_user)

        self.call_later(FALL_TIME, self._after_user_move)

    def _let_it_fall(self, piece, steps):
        """Move gems as if falling."""
        coord_x, old_cy = piece.orig_pos
        new_cy = old_cy - ARISTA * steps
        piece.do(actions.MoveTo((coord_x, new_cy), FALL_TIME))
        piece.orig_pos = coord_x, new_cy
        self.call_later(FALL_TIME, self.ph.piece_fall, piece, steps)

    def _after_user_move(self):
        """Finish the user movement."""
        # does the random-filled board has matches?
        m = self.ph.check_matches()
        if m:
            self._matches(m, from_user=False)
        else:
            self.falling = False
            # and there's any chance in the future?
            future = self.ph.future_guess()
            print "future", future
            if not future:
                self.game_over()

    def call_later(self, dt, func, *args, **kwargs):
        """Handy method to call later a function."""
        def f(dt, args, kwargs):
            """Unschedule the scheduled function."""
            self.unschedule(f)
            func(*args, **kwargs)
        self.schedule_interval(f, dt, args, kwargs)

    def game_over(self):
        """Sound and evaporation."""
        self.ended = True
        sounds.play("game_over")

        for x in range(8):
            for y in range(8):
                p = self.ph.get_piece(x, y)
                p.do(cocos.actions.FadeOut(1.5))

        # get the info
        self.do(cocos.actions.Rotate(360*2, 3) |
                cocos.actions.ScaleTo(10, 3) )



class MouseManager(cocos.layer.Layer):
    """Manager class for the mouse."""

    is_event_handler = True
    def __init__(self, piece_manager, board):
        super(MouseManager, self).__init__()
        self.pm = piece_manager
        self.board = board
        self.holding = None
        self.dragging = False

    def on_mouse_drag(self, coord_x, coord_y, delta_x, delta_y, buttons, mods):
        """User is dragging the cursor."""
        coord_x, coord_y = director.get_virtual_coordinates(coord_x, coord_y)

        if self.holding is None:
            slot = self.board.get_slot_from_cords(coord_x, coord_y)
            piece = self.pm.ph.get_piece(*slot)
            if piece is None:
                # trying to get a blowed up one
                return
            self.holding = (slot, piece)
            self.pm.hand_piece_select(piece)
            self.dragging = True

        # only move the piece if not selected it stationary
        if self.dragging:
            _, piece = self.holding
            self.pm.hand_piece_move(piece, (coord_x, coord_y))

    def on_mouse_release(self, coord_x, coord_y, buttons, modifiers):
        """User released a mouse button."""
        if self.pm.swapping or self.pm.ended or self.pm.paused:
            return
        coord_x, coord_y = director.get_virtual_coordinates(coord_x, coord_y)

        new_slot = self.board.get_slot_from_cords(coord_x, coord_y)
        if new_slot is None:
            if self.holding is not None:
                # release it if click outside the board
                (old_slot, old_piece) = self.holding
                self.pm.hand_piece_reset(old_piece)
                self.holding = None
            return

        new_sx, new_sy = new_slot
        new_piece = self.pm.ph.get_piece(new_sx, new_sy)
        if new_piece is None:
            # trying to get or get into a blowed up one
            return

        if self.holding is None:
            self.pm.hand_piece_select(new_piece)
            self.holding = (new_slot, new_piece)
            return

        (old_slot, old_piece) = self.holding
        old_sx, old_sy = old_slot
        if old_sx == new_sx and old_sy == new_sy:
            # we're in the same slot, unselect it
            self.pm.hand_piece_reset(old_piece)
            self.holding = None
            self.dragging = False
            return

        diff = abs(new_sx - old_sx) + abs(new_sy - old_sy)
        if diff > 1:
            # too far, reset old one...
            (old_slot, old_piece) = self.holding
            self.pm.hand_piece_reset(old_piece)
            self.holding = None

            # .. and hold a new one if the previous one wasn't dragged
            if not self.dragging:
                self.pm.hand_piece_select(new_piece)
                self.holding = (new_slot, new_piece)
            else:
                self.dragging = False

            return

        # move the original one
        self.pm.hand_piece_switch(old_piece, new_piece)
        self.holding = None
        self.dragging = False


class KeyboardManager(cocos.layer.Layer):
    """Handles the keyboard."""

    is_event_handler = True

    def __init__(self, game_manager):
        self.gm = game_manager
        cocos.layer.Layer.__init__(self)

    def on_key_press (self, key, modifiers):
        if key == keyboard.ESCAPE:
            self.gm.key_escape()
            return True

        if key in (keyboard.PAUSE, keyboard.P):
            self.gm.key_pause()
            return True

#        print "Key Pressed:", key, modifiers


class GameTiming(object):
    """Keeps the time of the game."""
    def __init__(self):
        self._start = time.time()
        self._is_paused = False
        self._time_paused = None

    def pause(self):
        """Stops counting the game time."""
        if self._is_paused:
            raise RuntimeError("Tried to pause while paused.")
        self._time_paused = time.time()
        self._is_paused = True

    def resume(self):
        """Starts counting game time again."""
        if not self._is_paused:
            raise RuntimeError("Tried to resume while not paused.")
        elapsed = time.time() - self._time_paused
        self._start += elapsed
        self._is_paused = False

    @property
    def time(self):
        """Returns the elapsed time."""
        if self._is_paused:
            t = self._time_paused - self._start
        else:
            t = time.time() - self._start
        return t


class GameManager(object):
    """Class that handles the games."""

    def __init__(self):
        self.playing = None
        self.timing = None

    def start(self, direct_play):
        if direct_play:
            m = getattr(self, 'play_' + direct_play, None)
            if m is None:
                print _("No such game!")
                exit()
            sc = m()
        else:
            sc = scene.Scene()
            mm = MainMenu(self)
            sc.add(mm, z=1)
        return sc

    def play_classic(self):
        """Classic game."""
        self.timing = GameTiming()
        sc = scene.Scene(BackgroundLayer())
        board = Board()
        sc.add(board, z=0)
        st = stats.Stats(self)
        self.pm = PieceManager(board, st)
        sc.add(self.pm, z=1)
        sc.add(MouseManager(self.pm, board), z=2)
        sc.add(KeyboardManager(self), z=2)
        sc.add(st, z=3)

        self.playing = sc
        return sc

    @property
    def game_time(self):
        return self.timing.time

    def continue_game(self):
        """Keep playing the game."""
        if not self.pm.paused:
            self.timing.resume()
        director.replace(self.playing)

    def key_escape(self):
        """Go back to main menu, but keeping the game."""
        if not self.pm.paused:
            self.timing.pause()
        if self.pm.ended:
            self.playing = None
        self.main_menu()

    def key_pause(self):
        """Pause the game."""
        now_paused = not self.pm.paused
        if now_paused:
            self.timing.pause()
        else:
            self.timing.resume()
        self.pm.set_pause(now_paused)

    def main_menu(self):
        sc = scene.Scene()
        mm = MainMenu(self, self.playing)
        sc.add(mm, z=1)
        director.replace(sc)
